import jwtDecode from 'jwt-decode';
import { extractQuerystringParameters } from '../utils/url';
export function getRealmUrl(realm, authServerUrl) {
  if (typeof authServerUrl === 'undefined') {
    return undefined;
  }

  if (authServerUrl.charAt(authServerUrl.length - 1) === '/') {
    return authServerUrl + 'realms/' + encodeURIComponent(realm);
  } else {
    return authServerUrl + '/realms/' + encodeURIComponent(realm);
  }
}
export function setupOidcEndoints({
  oidcConfiguration,
  realm,
  authServerUrl
}) {
  if (!oidcConfiguration) {
    if (!realm) {
      throw new Error('Missing realm');
    }

    return {
      authorize: function authorize() {
        return getRealmUrl(realm, authServerUrl) + '/protocol/openid-connect/auth';
      },
      token: function token() {
        return getRealmUrl(realm, authServerUrl) + '/protocol/openid-connect/token';
      },
      logout: function logout() {
        return getRealmUrl(realm, authServerUrl) + '/protocol/openid-connect/logout';
      },
      register: function register() {
        return getRealmUrl(realm, authServerUrl) + '/protocol/openid-connect/registrations';
      },
      userinfo: function userinfo() {
        return getRealmUrl(realm, authServerUrl) + '/protocol/openid-connect/userinfo';
      }
    };
  }

  return {
    authorize: function authorize() {
      return oidcConfiguration.authorization_endpoint;
    },
    token: function token() {
      return oidcConfiguration.token_endpoint;
    },
    logout: function logout() {
      if (!oidcConfiguration.end_session_endpoint) {
        throw 'Not supported by the OIDC server';
      }

      return oidcConfiguration.end_session_endpoint;
    },
    register: function register() {
      throw 'Redirection to "Register user" page not supported in standard OIDC mode';
    },
    userinfo: function userinfo() {
      if (!oidcConfiguration.userinfo_endpoint) {
        throw 'Not supported by the OIDC server';
      }

      return oidcConfiguration.userinfo_endpoint;
    }
  };
}
export function decodeToken(str) {
  return jwtDecode(str);
}
export function parseCallbackParams(paramsString, supportedParams) {
  const params = extractQuerystringParameters(paramsString);
  const [otherParams, oAuthParams] = Object.keys(params).reduce(([oParams, oauthParams], key) => {
    if (supportedParams.includes(key)) {
      oauthParams.set(key, params[key]);
    } else {
      oParams.add("".concat(key, "=").concat(params[key]));
    }

    return [oParams, oauthParams];
  }, [new Set(), new Map()]);
  return {
    paramsString: Array.from(otherParams.values()).join('&'),
    oauthParams: Object.fromEntries(oAuthParams.entries())
  };
}
export function isKeycloakConfig(config) {
  return !!config && typeof config !== 'string';
}
export async function fetchJSON(url, token) {
  const jsonRes = await fetch(url, {
    headers: {
      Accept: 'application/json',
      Authorization: token ? "Bearer ".concat(token) : ''
    }
  });
  return await jsonRes.json();
}
//# sourceMappingURL=utils.js.map